#include <cmath>
#include "color.h"
#include "colorcurve.h"
#include "mesh.h"
#include "shader.h"
#include "texture2d.h"

using namespace std;

namespace {

static const char fragment_src[] =
	"uniform sampler2D texture;\n"
	"uniform sampler1D curve;\n"
	"uniform float peak;\n"
	"uniform float brightness;\n"
	"varying vec2 texcoord;\n"
	"void main()\n"
	"{\n"
	"	vec4 sample = texture2D(texture, texcoord);\n"
	"	float maxc = max(sample.r, max(sample.g, sample.b));\n"
	"	if(maxc>1.0-peak)\n"
	"	{\n"
	"		vec3 saturated = sample.rgb/maxc;\n"
	"		if(maxc>1.0+peak)\n"
	"		{\n"
	"			sample.rgb = mix(vec3(1.0), saturated, 1.0/pow(brightness, maxc-1.0-peak));\n"
	"		}\n"
	"		else\n"
	"		{\n"
	"			float x = (1.0+peak-maxc)/(2.0*peak);\n"
	"			sample.rgb = saturated.rgb*(1.0-peak+(1.0-x*x)*peak);\n"
	"		}\n"
	"	}\n"
	"	gl_FragColor = vec4(texture1D(curve, sample.r).r, texture1D(curve, sample.g).r, texture1D(curve, sample.b).r, sample.a);\n"
	"}";

}

namespace Msp {
namespace GL {

ColorCurve::ColorCurve():
	quad(get_fullscreen_quad())
{
	shprog.attach_shader(get_fullscreen_vertex_shader());
	shprog.attach_shader_owned(new FragmentShader(fragment_src));
	shprog.bind_attribute(get_component_type(VERTEX2), "vertex");
	shprog.link();

	shdata.uniform("texture", 0);
	shdata.uniform("curve", 1);

	curve.storage(LUMINANCE, 256);
	curve.set_min_filter(LINEAR);
	curve.set_wrap(CLAMP_TO_EDGE);

	set_peak(0.2);
	set_brightness(1.5);
	set_linear();
}

void ColorCurve::set_peak(float p)
{
	if(p<0 || p>1)
		throw invalid_argument("ColorCurve::set_peak");
	shdata.uniform("peak", p);
}

void ColorCurve::set_brightness(float b)
{
	if(b<1)
		throw invalid_argument("ColorCurve::set_brightness");
	shdata.uniform("brightness", b);
}

void ColorCurve::set_gamma(float g)
{
	if(g<0.1 || g>10)
		throw invalid_argument("ColorCurve::set_gamma");

	unsigned char curve_data[256];
	for(unsigned i=0; i<256; ++i)
		curve_data[i] = pow(i/255.0f, 1/g)*255+0.5f;
	curve.image(0, LUMINANCE, UNSIGNED_BYTE, curve_data);
}

void ColorCurve::set_srgb()
{
	unsigned char curve_data[256];
	curve_data[0] = 0;
	for(unsigned i=1; i<256; ++i)
		curve_data[i] = to_srgb(i/255.0f)*255+0.5f;
	curve.image(0, LUMINANCE, UNSIGNED_BYTE, curve_data);
}

void ColorCurve::set_linear()
{
	unsigned char curve_data[256];
	for(unsigned i=0; i<256; ++i)
		curve_data[i] = i;
	curve.image(0, LUMINANCE, UNSIGNED_BYTE, curve_data);
}

void ColorCurve::render(const Texture2D &color_buf, const Texture2D &)
{
	Bind _bind_shader(shprog);
	shdata.apply();
	Bind _bind_mesh(quad);
	Bind _bind_tex(color_buf);
	Bind _bind_curve(curve, 1);
	quad.draw();
}

} // namespace GL
} // namespace Msp
