#ifndef MSP_GL_PROGRAMCOMPILER_H_
#define MSP_GL_PROGRAMCOMPILER_H_

#include <set>
#include "programparser.h"
#include "programsyntax.h"

namespace Msp {
namespace GL {

class Program;

class ProgramCompiler
{
private:
	struct Visitor: ProgramSyntax::TraversingVisitor
	{
		ProgramSyntax::Stage *stage;

		Visitor();

		virtual void apply(ProgramSyntax::Stage &);
	};

	struct Formatter: Visitor
	{
		std::string formatted;
		unsigned indent;
		bool parameter_list;
		bool else_if;

		Formatter();

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::Literal &);
		virtual void visit(ProgramSyntax::ParenthesizedExpression &);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::MemberAccess &);
		virtual void visit(ProgramSyntax::UnaryExpression &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::FunctionCall &);
		virtual void visit(ProgramSyntax::ExpressionStatement &);
		virtual void visit(ProgramSyntax::Layout &);
		virtual void visit(ProgramSyntax::StructDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
		virtual void visit(ProgramSyntax::FunctionDeclaration &);
		virtual void visit(ProgramSyntax::Conditional &);
		virtual void visit(ProgramSyntax::Iteration &);
		virtual void visit(ProgramSyntax::Return &);
	};

	struct VariableResolver: Visitor
	{
		std::vector<ProgramSyntax::Block *> blocks;
		ProgramSyntax::StructDeclaration *type;
		bool anonymous;

		VariableResolver();

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::MemberAccess &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::StructDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
	};

	struct InterfaceGenerator: Visitor
	{
		std::string in_prefix;
		std::string out_prefix;
		unsigned scope_level;
		std::map<std::string, ProgramSyntax::VariableDeclaration *> iface_declarations;
		bool remove_node;
		std::list<ProgramSyntax::Node *> insert_nodes;

		InterfaceGenerator();

		static std::string get_out_prefix(ProgramSyntax::StageType);
		virtual void apply(ProgramSyntax::Stage &);
		virtual void visit(ProgramSyntax::Block &);
		std::string change_prefix(const std::string &, const std::string &) const;
		bool generate_interface(ProgramSyntax::VariableDeclaration &, const std::string &, const std::string &);
		void insert_assignment(const std::string &, ProgramSyntax::Expression *);
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::Passthrough &);
	};

	struct VariableRenamer: Visitor
	{
		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
	};

	struct UnusedVariableLocator: Visitor
	{
		std::set<ProgramSyntax::Node *> unused_nodes;
		std::map<ProgramSyntax::VariableDeclaration *, ProgramSyntax::Node *> aggregates;
		ProgramSyntax::Node *aggregate;
		std::map<ProgramSyntax::VariableDeclaration *, ProgramSyntax::Node *> assignments;
		bool assignment;
		bool record_target;
		ProgramSyntax::VariableDeclaration *assignment_target;
		bool indeterminate_target;
		bool self_referencing;

		UnusedVariableLocator();

		virtual void visit(ProgramSyntax::VariableReference &);
		virtual void visit(ProgramSyntax::MemberAccess &);
		virtual void visit(ProgramSyntax::BinaryExpression &);
		virtual void visit(ProgramSyntax::ExpressionStatement &);
		virtual void visit(ProgramSyntax::StructDeclaration &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
		virtual void visit(ProgramSyntax::InterfaceBlock &);
	};

	struct NodeRemover: Visitor
	{
		std::set<ProgramSyntax::Node *> to_remove;

		virtual void visit(ProgramSyntax::Block &);
		virtual void visit(ProgramSyntax::VariableDeclaration &);
	};

	ProgramParser parser;
	ProgramSyntax::Module *module;

public:
	ProgramCompiler();

	void compile(const std::string &);
	void compile(IO::Base &);
	void add_shaders(Program &);

private:
	void process();
	void generate(ProgramSyntax::Stage &);
	bool optimize(ProgramSyntax::Stage &);
	static void inject_block(ProgramSyntax::Block &, const ProgramSyntax::Block &);
	template<typename T>
	static void apply(ProgramSyntax::Stage &);
	std::string create_source(ProgramSyntax::Stage &);
};

} // namespace GL
} // namespace Msp

#endif
